/*----------------------------------------------------------------------
  File    : xptlview.c
  Contents: Prototype-less cluster set visualization program for X11
  Author  : Christian Borgelt
  History : 2007.01.18 file created from file xbcview.c
            2007.01.23 reading of stored number of objects added
            2007.01.24 optional drawing of neighborhood graph added
            2007.01.25 input widget for marker border color added
            2013.07.22 adapted to definitions ATTID, VALID, TPLID etc.
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <assert.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Core.h>
#include <X11/Shell.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Viewport.h>
#include <X11/Xaw/List.h>
#include <X11/Xaw/AsciiText.h>
#include "menu.h"
#include "dialog.h"
#include "fselect.h"
#include "xshades.h"
#ifndef AS_READ
#define AS_READ
#endif
#ifndef AS_WRITE
#define AS_WRITE
#endif
#ifndef AS_PARSE
#define AS_PARSE
#endif
#include "attset.h"
#ifndef TAB_READ
#define TAB_READ
#endif
#include "table.h"
#include "ptlset.h"

#if ATTID_MAX > INT_MAX
#error "xbcview requires ATTID_MAX <= INT_MAX"
#endif
#if VALID_MAX > INT_MAX
#error "xbcview requires VALID_MAX <= INT_MAX"
#endif
#if DIMID_MAX > INT_MAX
#error "xbcview requires DIMID_MAX <= INT_MAX"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
/* --- extensions --- */
#define MINXEXT      120        /* minimal x-extension of window */
#define MINYEXT       90        /* minimal y-extension of window */
#define INITXEXT     400        /* initial x-extension of window */
#define INITYEXT     400        /* initial y-extension of window */
#define LISTWD       150        /* minimal list width */
#define LISTHT       120        /* minimal list width */
#define FMTCHRLEN     80        /* length of format character strings */

/* --- error codes --- */
#define E_NONE         0        /* no error */
#define E_NOMEM      (-1)       /* not enough memory */
#define E_FOPEN      (-2)       /* cannot open file */
#define E_FREAD      (-3)       /* read error on file */
#define E_FWRITE     (-4)       /* write error on file */
#define E_PARSE      (-5)       /* initialization failed */
#define E_CREATE     (-6)       /* cannot create cluster set */
#define E_ATTCNT     (-7)       /* too few numeric attributes */
#define E_DATA       (-8)       /* cluster set must be loaded first */
#define E_TPLCNT     (-9)       /* mismatch in number of records */
#define E_INIT      (-10)       /* initialization failed */
#define E_WIDGET    (-11)       /* widget creation failed */
#define E_DIALOG    (-12)       /* cannot create dialog box */
#define E_COLOR     (-13)       /* color allocation failed */
#define E_GRAPH     (-14)       /* cannot create graphics context */
#define E_DLOAD     (-15)       /* loading data failed */

#define FMT_MAX      256        /* max. len. of error message format */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* -- data format description -- */
  int    names;                 /* column names in first record */
  char   recseps[FMTCHRLEN+1];  /* record separators */
  char   fldseps[FMTCHRLEN+1];  /* field separators */
  char   blanks [FMTCHRLEN+1];  /* blank characters */
  char   nullchs[FMTCHRLEN+1];  /* null value characters */
  char   comment[FMTCHRLEN+1];  /* comment characters */
} DATAFMT;                      /* (data format description) */

typedef struct {                /* --- range of values --- */
  double min, max;              /* minimum and maximum value */
} RANGE;                        /* (range of values) */

typedef struct {                /* -- selected attributes -- */

  ATTID  h_att;                 /* attribute for horizontal direction */
  RANGE  h_rng;                 /* horizontal range of values */
  ATTID  v_att;                 /* attribute for vertical   direction */
  RANGE  v_rng;                 /* vertical range of values */
} SELATT;                       /* (selected attributes) */

typedef struct {                /* --- error message data --- */
  const char *name, *type;      /* name and type of the error message */
  const char *class;            /* resource class of error message */
  const char *dflt;             /* default message */
  int        argc;              /* number of arguments */
} ERRMSG;                       /* (error message) */

/*----------------------------------------------------------------------
  Action Function Prototypes
----------------------------------------------------------------------*/
static void ld_ptlset (Widget w, XEvent *e, String *s, Cardinal *c);
static void ld_data   (Widget w, XEvent *e, String *s, Cardinal *c);
static void format    (Widget w, XEvent *e, String *s, Cardinal *c);
static void quit      (Widget w, XEvent *e, String *s, Cardinal *c);
static void atts      (Widget w, XEvent *e, String *s, Cardinal *c);
static void colors    (Widget w, XEvent *e, String *s, Cardinal *c);
static void marks     (Widget w, XEvent *e, String *s, Cardinal *c);
static void ngbors    (Widget w, XEvent *e, String *s, Cardinal *c);
static void redraw    (Widget w, XEvent *e, String *s, Cardinal *c);
static void about     (Widget w, XEvent *e, String *s, Cardinal *c);
static void dclick    (Widget w, XEvent *e, String *s, Cardinal *c);

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
#include "xptlview.rsc"         /* fallback resources */

static XtActionsRec actions[] = {
  { "ld_ptlset", ld_ptlset },   /* File    > Load Cluster Set... */
  { "ld_data",   ld_data   },   /* File    > Load Data... */
  { "format",    format    },   /* File    > Data Format... */
  { "quit",      quit      },   /* File    > Quit */
  { "atts",      atts      },   /* Actions > Attributes... */
  { "colors",    colors    },   /* Actions > Colors... */
  { "marks",     marks     },   /* Actions > Markers... */
  { "ngbors",    ngbors    },   /* Actions > Neighbors... */
  { "redraw",    redraw    },   /* Actions > Redraw */
  { "about",     about     },   /* Help    > About XPtlView... */
  { "dclick",    dclick    },   /* double click in attribute list */
  { "db_focus",  db_focus  },   /* set focus to input gadget */
  { "db_next",   db_next   },   /* switch to next input gadget */
  { "db_close",  db_close  } }; /* close dialog box */

/* --- error messages --- */
static const ERRMSG errmsgs[] = {   /* error message data */
  /* E_NONE      0 */  { "error",   "none",    "XPtlView.Error",
                         "no error",                               0 },
  /* E_NOMEM    -1 */  { "error",   "nomem",   "XPtlView.Error",
                         "not enough memory",                      0 },
  /* E_FOPEN    -2 */  { "warning", "fopen",   "XPtlView.Warning",
                         "cannot open file:\n%s",                  1 },
  /* E_FREAD    -3 */  { "warning", "fread",   "XPtlView.Warning",
                         "read error on file:\n%s",                1 },
  /* E_FWRITE   -4 */  { "warning", "fwrite",  "XPtlView.Warning",
                         "write error on file:\n%s",               1 },
  /* E_PARSE    -5 */  { "warning", "parse",   "XPtlView.Warning",
                         "parse error on file:\n%s\n(see terminal "
                         "output for more information)",           1 },
  /* E_CREATE   -6 */  { "warning", "create",  "XPtlView.Error",
                         "cannot create cluster set",              0 },
  /* E_ATTCNT   -7 */  { "warning", "attcnt",  "XPtlView.Error",
                         "no numeric attributes",                  0 },
  /* E_DATA     -8 */  { "warning", "data",    "XPtlView.Error",
                         "a cluster set must be loaded first",     0 },
  /* E_TPLCNT   -9 */  { "warning", "misfld",  "XPtlView.Error",
                         "file %s:\nmismatch in number of records",1 },
  /* E_INIT    -10 */  { "error",   "init",    "XPtlView.Error",
                         "initialization failed",                  0 },
  /* E_WIDGET  -11 */  { "error",   "widget",  "XPtlView.Error",
                         "widget creation failed",                 0 },
  /* E_DIALOG  -12 */  { "warning", "dialog",  "XPtlView.Warning",
                         "cannot create dialog box",               0 },
  /* E_COLOR   -13 */  { "warning", "color",   "XPtlView.Warning",
                         "color allocation failed",                0 },
  /* E_GRAPH   -14 */  { "error",   "gc",      "XPtlView.Error",
                         "cannot create graphics context",         0 },
  /* E_DLOAD   -15 */  { "warning", "data",    "XBCView.Error",
                         "error loading data:\n%s",                1 },
  /*           -16 */  { "error", "unknown",   "XBCView.Error",
                         "unknown error",                          0 }};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static XtAppContext appctx;     /* X11 application context */
static Display *display;        /* display connected to */
static int     screen;          /* screen of the display */
static Window  window;          /* main window */
static Atom    wm_delwin;       /* delete window atom of window mgr. */
static char    fmt[FMT_MAX];    /* format and buffer for messages */
static char    buf[PATH_MAX +FMT_MAX];

/* --- main widgets and menu --- */
static Widget  w_top;           /* top level widget */
static Widget  w_main;          /* main window widget */
static Widget  w_view;          /* viewport widget */
static MENU    *menu;           /* pull down menu */
static GC      gc;              /* graphics context for drawing */

/* --- property sheets --- */
static PSHEET  *psh_fmt    = NULL;  /* data format */
static PSHEET  *psh_colors = NULL;  /* color    parameters */
static PSHEET  *psh_marks  = NULL;  /* marker   parameters */
static PSHEET  *psh_ngbors = NULL;  /* neighbor parameters */

/* --- dialog box widgets --- */
static Widget  db_atts = NULL;  /* 'Attributes...' dialog box */
static Widget  di_h_att;        /* attribute list for horizontal */
static Widget  di_h_min;        /* input gadgets for minimal */
static Widget  di_h_max;        /* and maximal horizontal value */
static Widget  di_v_att;        /* attribute list for vertical */
static Widget  di_v_min;        /* input gadgets for minimal */
static Widget  di_v_max;        /* and maximal vertical value */

/* --- color and marker information --- */
static PIXEL   black;           /* black, white, and grey */
static PIXEL   white;           /* for markers on colored shades */
static PIXEL   grey;
static SHADES  *shades  = NULL; /* color shades for normal dists. */
static SHADES  *datcols = NULL; /* colors for data points/markers */

static struct {                 /* --- parameters --- */
  int    nbcnt;                 /* number of neighbors */
  int    graph;                 /* flag for neighborhood graph */
  int    size;                  /* size of data markers */
  int    border;                /* color of data marker borders */
  int    shdbase;               /* shade base */
  int    shdcnt;                /* number of shades */
  int    colcnt;                /* number of colors */
  double coloff;                /* color offset (in degrees) */
} parblk = { 0, 0, 6, 0, SHD_WHITE, 32, 1, 0.0 };

/* --- attribute information --- */
ATTSET  *attset  = NULL;        /* attribute set */
CCHAR   **names  = NULL;        /* attribute names */
RANGE   *ranges  = NULL;        /* ranges of attribute values */
ATTID   *map_fwd = NULL;        /* map indices to attset identifiers */
ATTID   *map_inv = NULL;        /* map attset identifiers to indices */
int     attcnt   = 0;           /* number of (metric) attributes */
SELATT  selatt   =              /* attribute selection information */
  { 0, { 0.0, 0.0 }, 0, { 0.0, 0.0 } };

/* --- cluster and data information --- */
TABREAD *tread   = NULL;        /* table file reader */
DATAFMT datafmt  =              /* data format information */
  { 1, "\\n", " ,\\t", " \\t\\r", "?", "#" };
PTLSET  *ptlset  = NULL;        /* prototype-less cluster set */
TABLE   *ptltab  = NULL;        /* cluster set table */
DISTMAT *distmat = NULL;        /* distance matrix for data table */
TABLE   *table   = NULL;        /* additional data table */
double  *scales  = NULL;        /* scaling factors for normalization */
double  *offs    = NULL;        /* offsets for normalization */
double  *dists   = NULL;        /* vector of distances/m.s. degrees */
int     *nbids   = NULL;        /* vector of neighbor identifiers */

/* --- error message --- */
char    ldmsg[2*AS_MAXLEN+64];  /* buffer for an error message */

/*----------------------------------------------------------------------
  Error Reporting Functions
----------------------------------------------------------------------*/

static void dcb_nop (Widget widget, XtPointer client, XtPointer call)
{                               /* --- close dialog box (callback) */
  while (!XtIsTransientShell(widget))
    widget = XtParent(widget);  /* find the shell widget */
  XtPopdown(widget);            /* close dialog box */
  mn_enable(menu, MN_MENU, 1);  /* and enable menu bar */
}  /* dcb_nop() */

/*--------------------------------------------------------------------*/

static void error (int code, ...)
{                               /* --- print error message */
  va_list      va;              /* list of variable arguments */
  int          k;               /* minimal error code */
  const ERRMSG *msg;            /* error message */
  String       argv[8];         /* variable argument vector */
  Cardinal     argc;            /* number of arguments */

  k = 1-(int)(sizeof(errmsgs)/sizeof(*errmsgs));
  if (code < k) code = k;       /* check error code against minimum */
  if (code > 0) code = E_DLOAD; /* check for a data load error */
  msg = errmsgs -code;          /* get error message data */
  if (!msg->name) msg = errmsgs -k;
  va_start(va, code);           /* get list of variable arguments */
  for (argc = 0; (int)argc < msg->argc; argc++)
    argv[argc] = va_arg(va, String);
  va_end(va);                   /* get message arguments */
  if (strcmp(msg->name, "error") == 0)  /* if name is 'error' */
    XtAppErrorMsg(appctx, msg->name, msg->type, msg->class,
      msg->dflt, argv, &argc);  /* print error message */
  else {                        /* if name is 'warning' */
    XtAppGetErrorDatabaseText(appctx, msg->name, msg->type,
      msg->class, msg->dflt, fmt, FMT_MAX, NULL);
    va_start(va, code);         /* list of variable arguments */
    vsprintf(buf, fmt, va);     /* get and format message */
    va_end(va);                 /* end variable argument evaluation */
    mn_enable(menu, MN_MENU,0); /* disable menu bar and show dialog */
    if (db_alert(w_top, dcb_nop, buf) != 0) {
      mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
      XtAppWarningMsg(appctx, msg->name, msg->type, msg->class,
        msg->dflt, argv, &argc);/* if the alert dialog box */
    }                           /* cannot be shown, */
  }                             /* print a warning message */
}  /* error() */

/*----------------------------------------------------------------------
  Load Functions
----------------------------------------------------------------------*/

static void cleanup (int all)
{                               /* --- clean up after a load error */
  if (tread)   { trd_delete(tread,  1); tread   = NULL; }
  if (all <  0) return;
  if (table)   { tab_delete(table,  0); table   = NULL; }
  if (all <= 0) return;
  if (distmat) { dm_delete(distmat);    distmat = NULL; }
  if (ptltab)  { tab_delete(ptltab, 0); ptltab  = NULL; }
  if (ptlset)  { ptl_delete(ptlset);    ptlset  = NULL; }
  if (attset)  { as_delete(attset);     attset  = NULL; }
  if (dists)   { free(dists);           dists   = NULL; }
  if (nbids)   { free(nbids);           nbids   = NULL; }
  if (ranges)  { free(ranges);          ranges  = NULL; }
  if (map_fwd) { free(map_fwd);         map_fwd = NULL; }
  if (names)   { free((void*)names);    names   = NULL; }
  attcnt = 0;                   /* free memory and clear variables */
}  /* cleanup() */

/*--------------------------------------------------------------------*/

static int lderr (int code, int all, const char *msg, SCANNER *scan)
{                               /* --- handle a load error */
  if (scan) scn_delete(scan,1); /* delete the scanner */
  if (!msg) ldmsg[0] = 0;       /* note the error message */
  else strncpy(ldmsg, msg, sizeof(ldmsg));
  ldmsg[sizeof(ldmsg)-1] = 0;   /* terminate the error message */
  cleanup(all);                 /* clear all/data variables */
  return code;                  /* return the error code */
}  /* lderr() */

/*--------------------------------------------------------------------*/

static int load_ptlset (const char *fname)
{                               /* --- load a p.t.-less cluster set */
  ATTID   k, m;                 /* loop variables for attributes */
  DIMID   r, n;                 /* loop variables for rows/tuples */
  int     type;                 /* attribute type */
  SCANNER *scan;                /* scanner (to parse cluster set) */
  ATT     *att;                 /* to traverse the attributes */

  /* --- parse domain descriptions --- */
  scan = scn_create();          /* create a scanner */
  if (!scan) return E_NOMEM;    /* for the classifier/cluster set */
  cleanup(1);                   /* clean up existing variables */
  if (scn_open(scan, NULL, fname) != 0)
    return lderr(E_FOPEN, 1, fname, scan);
  attset = as_create("domains", att_delete);
  if (!attset) return lderr(E_NOMEM, 1, NULL, scan);
  if (as_parse(attset, scan, AT_ALL, 1) != 0)
    return lderr(E_PARSE, 1, fname, scan);

  /* --- parse cluster set --- */
  ptlset = ptl_parse(scan);     /* parse a prototype-less cluster set */
  if (!ptlset || !scn_eof(scan, 1))
    return lderr(E_PARSE, 1, fname, scan);
  scn_delete(scan, 1);          /* delete the scanner */

  /* --- create the attribute vectors --- */
  m = as_attcnt(attset);        /* get the number of attributes */
  for (k = attcnt = 0; k < m; k++) {
    att  = as_att(attset, k);   /* traverse the attributes */
    type = att_type(att);       /* and check their types */
    if      (type == AT_INT) att_convert(att, AT_FLT, NULL);
    else if (type != AT_FLT) continue;
    attcnt++;                   /* make numeric atts. real-valued */
  }                             /* count the numeric attributes */
  if (attcnt < 1)               /* and check for at least one */
    return lderr(E_ATTCNT, 1, fname, NULL);
  n = ptl_rowcnt(ptlset); r = n +ptl_clscnt(ptlset);
  dists   = (double*)malloc((2*(size_t)attcnt
                              +(size_t)r)     *sizeof(double));
  nbids   = (DIMID*) malloc( 2*(size_t)n      *sizeof(DIMID));
  names   = (CCHAR**)malloc(   (size_t)attcnt *sizeof(CCHAR*));
  ranges  = (RANGE*) malloc(   (size_t)attcnt *sizeof(RANGE));
  map_fwd = (ATTID*) malloc( 2*(size_t)attcnt *sizeof(ATTID));
  if (!dists || !nbids || !names || !ranges || !map_fwd)
    return E_NOMEM;             /* create the needed buffers */
  map_inv = map_fwd +attcnt;    /* create the attribute arrays */
  scales  = dists +r; offs = scales +attcnt;
  parblk.nbcnt = ptl_getobjc(ptlset);
  if (parblk.nbcnt < 0) parblk.nbcnt = 0;

  /* --- build cluster set and attribute information --- */
  selatt.h_att = selatt.v_att = -1;
  for (k = attcnt = 0; k < m; k++) {
    att  = as_att(attset, k);   /* traverse the attributes */
    type = att_type(att);       /* skip non-metric attributes */
    if (type != AT_FLT) continue;
    if      (selatt.h_att < 0)  /* note the first  metric attribute */
      selatt.h_att = attcnt;    /* for the horizontal direction */
    else if (selatt.v_att < 0)  /* note the second metric attribute */
      selatt.v_att = attcnt;    /* for the vertical direction */
    names  [attcnt] = att_name(att);
    map_fwd[attcnt] = k;        /* note the attribute name */
    map_inv[k]      = attcnt++; /* and  the attribute identifier */
  }
  if (selatt.v_att < 0)         /* handle one num. att. properly */
    selatt.v_att = selatt.h_att;
  return 0;                     /* return 'ok' */
}  /* load_ptlset() */

/*--------------------------------------------------------------------*/

static int load_data (const char *fname)
{                               /* --- load data table */
  ATTID  k;                     /* loop variable */
  int    r;                     /* result of a function call */
  int    m;                     /* table file read mode */
  ATT    *att;                  /* to traverse the attributes */
  RANGE  *rng;                  /* to traverse the ranges */
  DTFLT  f;                     /* buffer for a float value */
  double s;                     /* span of values */
  TABLE  *tab = NULL;           /* created table */

  /* --- load a data table --- */
  if (!ptlset) return E_DATA;   /* check for an attribute set */
  for (k = as_attcnt(attset); --k >= 0; )
    att_setmark(as_att(attset, k), 1);
  tab = tab_create("table", attset, tpl_delete);
  if (!tab) return E_NOMEM;     /* create a data table */
  tread = trd_create();         /* create a table file reader */
  if (!tread) { tab_delete(tab, 0); return E_NOMEM; }
  trd_allchs(tread, datafmt.recseps, datafmt.fldseps,
                    datafmt.blanks,  datafmt.nullchs, datafmt.comment);
  if (trd_open(tread, NULL, fname) != 0) { /* open the input file */
    tab_delete(tab, 0); return lderr(E_FOPEN, -1, fname, NULL); }
  m = ((datafmt.names) ? AS_ATT : AS_DFLT) | AS_MARKED;
  r = tab_read(tab, tread, m);  /* load a data table */
  if (r < 0) {                  /* from the given file */
    strncpy(ldmsg, tab_errmsg(table, NULL, 0), sizeof(ldmsg));
    ldmsg[sizeof(ldmsg)-1] = 0; /* if an error occurred, */
    tab_delete(tab, 0);         /* copy the error message */
    return -r;                  /* delete the created table */
  }                             /* and abort the function */
  trd_delete(tread, 1);         /* read the table body and */
  tread = NULL;                 /* delete the table reader */

  /* --- assign the loaded table --- */
  if (ptltab) {                 /* if a data table is needed */
    if (table) tab_delete(table, 0);
    table = tab; }              /* set the new data table */
  else {                        /* if a cluster set table is needed */
    if (tab_tplcnt(tab) != ptl_rowcnt(ptlset)) {
      tab_delete(tab, 0); return E_TPLCNT; }
    for (k = 0; k < attcnt; k++) {
      k   = map_fwd[k];         /* traverse the numeric attributes */
      att = as_att(attset, k);  /* and normalize column if necessary */
      rng = ranges +k;          /* determine the range of values */
      f   = att_valmin(att)->f; /* get and set minimum value */
      rng->min = (f <= DTFLT_MIN) ? -INFINITY : (double)f;
      f   = att_valmax(att)->f; /* get and set maximum value */
      rng->max = (f >= DTFLT_MAX) ?  INFINITY : (double)f;
      s   = (DTFLT)(rng->max -rng->min);
      if      (s <  0) { rng->min  = 0; rng->max  = 1; }
      else if (s <= 0) { rng->min -= 1; rng->max += 1; }
      else       { rng->min -= 0.05 *s; rng->max += 0.05 *s; }
    }                           /* adapt to obtain final range */
    selatt.h_rng = ranges[selatt.h_att];
    selatt.v_rng = ranges[selatt.v_att];
    if (!ptl_getnorm(ptlset))   /* if unnormalized data, */
      ptltab = tab;             /* the table can be used directly */
    else {                      /* if normalized data */
      ptltab = tab_clone(tab, 0); /* clone the loaded table */
      for (k = 0; k < attcnt; k++)
        tab_colnorm(ptltab, map_fwd[k], 0, 1, scales+k, offs+k);
    }                           /* normalize the numeric columns */
    table = tab;                /* set new data table (coincides */
  }                             /* with the cluster set table) */

  /* --- compute distance matrix --- */
  if (distmat) { dm_delete(distmat); distmat = NULL; }
  if ((parblk.nbcnt > 0) && parblk.graph) {
    distmat = dm_create(tab_tplcnt(table), parblk.nbcnt);
    if (!distmat) return E_NOMEM;
    dm_fromtab(distmat, table); /* create a distance matrix */
  }                             /* and fill it from the table */
  return 0;                     /* return 'ok' */
}  /* load_data() */

/*----------------------------------------------------------------------
  Color Functions
----------------------------------------------------------------------*/

static void col_delete (void)
{                               /* --- delete color shades */
  if (datcols) { shd_delete(datcols); datcols = NULL; }
  if (shades)  { shd_delete(shades);  shades  = NULL; }
}  /* col_delete() */           /* delete all shades/colors */

/*--------------------------------------------------------------------*/

static int col_create (void)
{                               /* --- create color shades */
  int      clscnt;              /* number of clusters */
  Colormap colmap;              /* default colormap */

  if (!ptlset) return 0;        /* check for an existing cluster set */
  col_delete();                 /* delete existing shades/colors */
  clscnt = ptl_clscnt(ptlset);  /* note the number of colors */
  if (parblk.colcnt > 0) parblk.colcnt = clscnt;
  colmap = DefaultColormap(display, screen);
  shades = shd_create(display, colmap, parblk.coloff,  parblk.colcnt,
                                       parblk.shdbase, parblk.shdcnt);
  if (!shades) return -1;       /* create color shades for distribs. */
  datcols = shd_create(display, colmap, parblk.coloff, clscnt,
                                        SHD_BLACK,      1);
  if (!datcols) { col_delete(); return -1; }
  return 0;                     /* get colors for data points */
}  /* col_create() */           /* and return 'ok' */

/*----------------------------------------------------------------------
  Dialog Box Callback Functions
----------------------------------------------------------------------*/

static void dcb_ptlset (Widget widget, XtPointer ok, XtPointer fname)
{                               /* --- 'Load Cluster Set...' callback */
  int r;                        /* result of load_ptlset() */

  mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
  if (!ok) return;              /* if not ok pressed, abort */
  r = load_ptlset(fname);       /* load a set of clusters */
  if (r != 0) { error(r, fname); return; }
  if (col_create() != 0) {      /* get new color shades */
    error(E_COLOR); return; }   /* and marker colors */
  if ((ptrdiff_t)ok < 2)        /* if not called on startup, */
    redraw(NULL,NULL,NULL,NULL);/* redraw the window contents */
}  /* dcb_ptlset() */

/*--------------------------------------------------------------------*/

static void dcb_data (Widget widget, XtPointer ok, XtPointer fname)
{                               /* --- 'Load Data...' callback */
  int r;                        /* result of load_data() */

  mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
  if (!ok || !ptlset) return;   /* if not ok pressed, abort */
  r = load_data(fname);         /* load a data table */
  if (r) { error(r, (r > 0) ? ldmsg : fname); return; }
  redraw(NULL,NULL,NULL,NULL);  /* redraw the window contents */
}  /* dcb_data() */

/*--------------------------------------------------------------------*/

static void dcb_format (Widget widget, XtPointer ok, XtPointer call)
{                               /* --- 'Data Format...' callback */
  mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
  if (!ok) return;              /* if not ok pressed, abort */

  datafmt.names = (int)(ptrdiff_t)psh_getval(psh_fmt, "names");
  strncpy(datafmt.recseps, psh_getval(psh_fmt, "record"),  FMTCHRLEN);
  datafmt.recseps[FMTCHRLEN] = '\0';  /* get record separators */
  strncpy(datafmt.fldseps, psh_getval(psh_fmt, "field"),   FMTCHRLEN);
  datafmt.fldseps[FMTCHRLEN] = '\0';  /* get field separators */
  strncpy(datafmt.blanks,  psh_getval(psh_fmt, "blank"),   FMTCHRLEN);
  datafmt.blanks [FMTCHRLEN] = '\0';  /* get blank characters */
  strncpy(datafmt.nullchs, psh_getval(psh_fmt, "null"),    FMTCHRLEN);
  datafmt.nullchs[FMTCHRLEN] = '\0';  /* get null value characters */
  strncpy(datafmt.comment, psh_getval(psh_fmt, "comment"), FMTCHRLEN);
  datafmt.comment[FMTCHRLEN] = '\0';  /* get comment characters */
}  /* dcb_format() */

/*--------------------------------------------------------------------*/

static void dcb_atts (Widget widget, XtPointer ok, XtPointer call)
{                               /* --- 'Attributes...' callback */
  int    i;                     /* list index */
  String s;                     /* text in input gadgets */

  if (((ptrdiff_t)ok != 2)      /* if not to keep dialog open */
  && (((ptrdiff_t)ok != 1) || ((ptrdiff_t)call != 2))) {
    XtPopdown(db_atts);         /* close dialog box */
    mn_enable(menu, MN_MENU, 1);/* enable menu bar */
  }
  if ((ptrdiff_t)ok == 0) return;  /* if not ok pressed, abort */

  i = XawListShowCurrent(di_h_att)->list_index;
  if (i != XAW_LIST_NONE) {     /* if an attribute is selected, */
    selatt.h_att = i;           /* note it for the horizontal dir. */
    XtVaGetValues(di_h_min, XtNstring, &s, NULL);
    ranges[selatt.h_att].min = selatt.h_rng.min = atof(s);
    XtVaGetValues(di_h_max, XtNstring, &s, NULL);
    ranges[selatt.h_att].max = selatt.h_rng.max = atof(s);
  }                             /* get the range of values */
  i = XawListShowCurrent(di_v_att)->list_index;
  if (i != XAW_LIST_NONE) {     /* if an attribute is selected, */
    selatt.v_att = i;           /* note it for the vertical dir. */
    XtVaGetValues(di_v_min, XtNstring, &s, NULL);
    ranges[selatt.v_att].min = selatt.v_rng.min = atof(s);
    XtVaGetValues(di_v_max, XtNstring, &s, NULL);
    ranges[selatt.v_att].max = selatt.v_rng.max = atof(s);
  }                             /* get the range of values */
  redraw(NULL,NULL,NULL,NULL);  /* redraw the window contents */
}  /* dcb_atts() */

/*--------------------------------------------------------------------*/

static void dcb_colors (Widget widget, XtPointer ok, XtPointer call)
{                               /* --- 'Colors...' callback */
  int  clscnt;                  /* number of clusters */
  char buf[32];                 /* temporary buffer */

  mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
  if (!ok) return;              /* if not ok pressed, abort */
  clscnt = (ptlset) ? ptl_clscnt(ptlset) : 1;
  parblk.colcnt  = (psh_getval(psh_colors, "mode") == (void*)0)
                 ? clscnt : 0;  /* get shading mode */
  strncpy(buf, psh_getval(psh_colors, "coloff"), 32); buf[31] = '\0';
  parblk.coloff  = atof(buf);   /* get color offset (angle) */
  parblk.shdbase = (psh_getval(psh_colors, "base") == (void*)0)
                 ? SHD_BLACK : SHD_WHITE;  /* get shade base */
  strncpy(buf, psh_getval(psh_colors, "shdcnt"), 32); buf[31] = '\0';
  parblk.shdcnt  = atoi(buf);  /* get number of shades */
  if (parblk.shdcnt < 1)          parblk.shdcnt = 1;
  if (parblk.shdcnt > SHD_MAXCNT) parblk.shdcnt = SHD_MAXCNT;
  if (col_create() != 0)        /* get new color shades */
    error(E_COLOR);             /* and marker colors */
  redraw(NULL,NULL,NULL,NULL);  /* redraw the window contents */
}  /* dcb_colors() */

/*--------------------------------------------------------------------*/

static void dcb_marks (Widget widget, XtPointer ok, XtPointer call)
{                               /* --- 'Markers...' callback */
  mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
  if (!ok) return;              /* if not ok pressed, abort */
  parblk.size   = atoi(psh_getval(psh_marks, "size"));
  if (parblk.size <= 0) parblk.size = 1;
                                /* get the marker size */
  parblk.border = (psh_getval(psh_marks, "border") == (void*)0)
                ? 0 : 1;        /* get the marker color */
  redraw(NULL,NULL,NULL,NULL);  /* redraw the window contents */
}  /* dcb_marks() */

/*--------------------------------------------------------------------*/

static void dcb_ngbors (Widget widget, XtPointer ok, XtPointer call)
{                               /* --- 'Neighbors...' callback */
  TABLE *tab;                   /* table to work on */

  mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
  if (!ok) return;              /* if not ok pressed, abort */
  parblk.nbcnt   = atoi(psh_getval(psh_ngbors, "nbcnt"));
  if (parblk.nbcnt < 0)         /* get the number of neighbors */
    parblk.nbcnt = (ptlset) ? ptl_getobjc(ptlset) : 0;
  parblk.graph   = (psh_getval(psh_ngbors, "graph") == (void*)0)
                 ? 0 : 1;       /* get the neighbor graph flag */
  if (distmat) { dm_delete(distmat); distmat = NULL; }
  if ((parblk.nbcnt > 0) && parblk.graph) {
    tab = (table) ? table : ptltab;
    distmat = dm_create(tab_tplcnt(tab), parblk.nbcnt);
    if (!distmat) error(E_NOMEM);
    dm_fromtab(distmat, tab);   /* create a distance matrix */
  }                             /* and fill it from the table */
  redraw(NULL,NULL,NULL,NULL);  /* redraw the window contents */
}  /* dcb_ngbors() */

/*----------------------------------------------------------------------
  Menu Callback Functions
----------------------------------------------------------------------*/

static void mcb_ptlset (Widget widget, XtPointer client, XtPointer call)
{ ld_ptlset(NULL, NULL, NULL, NULL); }

static void mcb_data   (Widget widget, XtPointer client, XtPointer call)
{ ld_data(NULL, NULL, NULL, NULL); }

static void mcb_format (Widget widget, XtPointer client, XtPointer call)
{ format(NULL, NULL, NULL, NULL); }

static void mcb_quit   (Widget widget, XtPointer client, XtPointer call)
{ quit(NULL, NULL, NULL, NULL); }

static void mcb_colors (Widget widget, XtPointer client, XtPointer call)
{ colors(NULL, NULL, NULL, NULL); }

static void mcb_marks  (Widget widget, XtPointer client, XtPointer call)
{ marks(NULL, NULL, NULL, NULL); }

static void mcb_ngbors (Widget widget, XtPointer client, XtPointer call)
{ ngbors(NULL, NULL, NULL, NULL); }

static void mcb_atts   (Widget widget, XtPointer client, XtPointer call)
{ atts(NULL, NULL, NULL, NULL); }

static void mcb_redraw (Widget widget, XtPointer client, XtPointer call)
{ redraw(NULL, NULL, NULL, NULL); }

static void mcb_about  (Widget widget, XtPointer client, XtPointer call)
{ about(NULL, NULL, NULL, NULL); }

/*----------------------------------------------------------------------
  List Callback Function
----------------------------------------------------------------------*/

static void lcb_att (Widget widget, XtPointer client, XtPointer call)
{                               /* --- attribute list callback */
  int  i;                       /* list index */
  char buf[32];                 /* conversion buffer */

  i = ((XawListReturnStruct*)call)->list_index;
  if ((i == XAW_LIST_NONE) || !ranges)
    return;                     /* check selected list item */
  widget = ((ptrdiff_t)client == 0) ? di_h_min : di_v_min;
  sprintf(buf, "%g", ranges[i].min);
  XtVaSetValues(widget, XtNstring, buf, NULL);
  widget = ((ptrdiff_t)client == 0) ? di_h_max : di_v_max;
  sprintf(buf, "%g", ranges[i].max);
  XtVaSetValues(widget, XtNstring, buf, NULL);
}  /* lcb_att() */              /* set new range of values */

/*----------------------------------------------------------------------
  Action Functions
----------------------------------------------------------------------*/

static void ld_ptlset (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- File > Load Cluster Set... */
  const char *t;                /* buffer for dialog box title */

  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  XtVaGetValues(XtNameToWidget(mn_menubar(menu), "*ld_ptlset"),
    XtNlabel, &t, NULL);        /* get name of menu entry */
  if (fselect(w_top, t, NULL, NULL, dcb_ptlset) != 0) {
    error(E_DIALOG); mn_enable(menu, MN_MENU, 1); }
}  /* ld_ptlset() */            /* show a file selector box */

/*--------------------------------------------------------------------*/

static void ld_data (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- File > Load Data... */
  const char *t;                /* buffer for dialog box title */

  if (!ptlset) {                /* if no cluster set has been loaded, */
    error(E_DATA); return; }    /* no data file can be loaded */
  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  XtVaGetValues(XtNameToWidget(mn_menubar(menu), "*ld_data"),
    XtNlabel, &t, NULL);        /* get name of menu entry */
  if (fselect(w_top, t, NULL, NULL, dcb_data) != 0) {
    error(E_DIALOG); mn_enable(menu, MN_MENU, 1); }
}  /* ld_data() */              /* show a file selector box */

/*--------------------------------------------------------------------*/

static void format (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- File > Data Format... */
  if (!psh_fmt) {               /* if property sheet not created yet */
    psh_fmt = psh_create("fmtDlg", w_top, 1);
    if (!psh_fmt) return;       /* create a property sheet */
    if ((psh_addline(psh_fmt, "lnames")               != 0)
    ||  (psh_additem(psh_fmt, "names",   PSH_MENU, 0) != 0)
    ||  (psh_menuadd(psh_fmt, "names",   "no")        != 0)
    ||  (psh_menuadd(psh_fmt, "names",   "yes")       != 0)
    ||  (psh_addline(psh_fmt, "lrecord")              != 0)
    ||  (psh_additem(psh_fmt, "record",  PSH_EDIT, 0) != 0)
    ||  (psh_addline(psh_fmt, "lfield")               != 0)
    ||  (psh_additem(psh_fmt, "field",   PSH_EDIT, 0) != 0)
    ||  (psh_addline(psh_fmt, "lblank")               != 0)
    ||  (psh_additem(psh_fmt, "blank",   PSH_EDIT, 0) != 0)
    ||  (psh_addline(psh_fmt, "lnull")                != 0)
    ||  (psh_additem(psh_fmt, "null",    PSH_EDIT, 0) != 0)
    ||  (psh_addline(psh_fmt, "lcomment")             != 0)
    ||  (psh_additem(psh_fmt, "comment", PSH_EDIT, 0) != 0)) {
      psh_delete(psh_fmt); error(E_DIALOG); return; }
  }                             /* add dialog items */
  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  psh_setval(psh_fmt, "names",   (char*)(ptrdiff_t)datafmt.names);
  psh_setval(psh_fmt, "blank",   datafmt.blanks);
  psh_setval(psh_fmt, "field",   datafmt.fldseps);
  psh_setval(psh_fmt, "record",  datafmt.recseps);
  psh_setval(psh_fmt, "null",    datafmt.nullchs);
  psh_setval(psh_fmt, "comment", datafmt.comment);
  psh_handle(psh_fmt, dcb_format); /* fill input gadgets with current */
}  /* format() */                  /* values and handle dialog */

/*--------------------------------------------------------------------*/

static void quit (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- File > Quit */
  cleanup(1);                   /* clean up the variables and */
  col_delete();                 /* delete all colors and shades */
  if (w_top) XtDestroyApplicationContext(appctx);
  exit(0);                      /* terminate the program */
}  /* quit() */

/*--------------------------------------------------------------------*/

static void dclick (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- double click in attribute list */
  dcb_atts(db_atts, (void*)2, NULL);
}  /* dclick() */               /* call dialog box callback for 'ok' */

/*--------------------------------------------------------------------*/

static int listbox (Widget form, Widget left, const char *name)
{                               /* --- create a list box */
  Cardinal n;                   /* widget argument counter */
  Arg      args[16];            /* widget argument list */
  Widget   view, list;          /* viewport and list widget */
  Widget   min, max;            /* minimum and maximum widget */
  Widget   label;               /* buffer for label widgets */
  char     buf[16];             /* buffer for widget name */

  /* --- create list box --- */
  n = 0;                      /* build argument list */
  XtSetArg(args[n], XtNborderWidth, 0);                       n++;
  if (left) { XtSetArg(args[n], XtNfromHoriz,   left);        n++; }
  else      { XtSetArg(args[n], XtNleft,        XtChainLeft); n++; }
  XtSetArg(args[n], XtNtop,         XtChainTop);              n++;
  XtSetArg(args[n], XtNbottom,      XtChainTop);              n++;
  XtSetArg(args[n], XtNjustify,     XtJustifyLeft);           n++;
  sprintf(buf, "l%s", name);  /* create the widget name */
  label = XtCreateManagedWidget(buf, labelWidgetClass, form, args, n);
  if (!label) return -1;      /* create title widget */
  n = 0;                      /* build argument list */
  XtSetArg(args[n], XtNborderWidth, 1);                       n++;
  if (left) { XtSetArg(args[n], XtNfromHoriz,   left);        n++; }
  else      { XtSetArg(args[n], XtNleft,        XtChainLeft); n++; }
  XtSetArg(args[n], XtNfromVert,    label);                   n++;
  XtSetArg(args[n], XtNallowHoriz,  False);                   n++;
  XtSetArg(args[n], XtNallowVert,   True);                    n++;
  XtSetArg(args[n], XtNuseRight,    True);                    n++;
  XtSetArg(args[n], XtNtop,         XtChainTop);              n++;
  XtSetArg(args[n], XtNbottom,      XtChainBottom);           n++;
  sprintf(buf, "v%s", name);  /* create the widget name */
  view = XtCreateManagedWidget(buf, viewportWidgetClass,
                               form, args, n);
  if (!view) return -1;       /* create list viewport */
  n = 0;                      /* build argument list */
  XtSetArg(args[n], XtNstring,         "");                   n++;
  XtSetArg(args[n], XtNborderWidth,    0);                    n++;
  XtSetArg(args[n], XtNverticalList,   True);                 n++;
  XtSetArg(args[n], XtNdefaultColumns, 1);                    n++;
  XtSetArg(args[n], XtNforceColumns,   True);                 n++;
  sprintf(buf, "%s_att", name);  /* create the widget name */
  list = XtCreateManagedWidget(buf, listWidgetClass, view, args, n);
  if (!list) return -1;       /* create list widget */

  /* --- create input field for minimum --- */
  n = 0;                      /* build argument list */
  XtSetArg(args[n], XtNborderWidth, 0);                       n++;
  XtSetArg(args[n], XtNfromVert,    view);                    n++;
  if (left) { XtSetArg(args[n], XtNfromHoriz,   left);        n++; }
  else      { XtSetArg(args[n], XtNleft,        XtChainLeft); n++; }
  XtSetArg(args[n], XtNtop,         XtChainBottom);           n++;
  XtSetArg(args[n], XtNbottom,      XtChainBottom);           n++;
  XtSetArg(args[n], XtNjustify,     XtJustifyLeft);           n++;
  sprintf(buf, "l%s_min", name); /* create the widget name */
  label = XtCreateManagedWidget(buf, labelWidgetClass, form, args, n);
  if (!label) return -1;      /* create min label widget */
  n = 0;                      /* build argument list */
  XtSetArg(args[n], XtNborderWidth, 1);                       n++;
  XtSetArg(args[n], XtNeditType,    XawtextEdit);             n++;
  XtSetArg(args[n], XtNfromHoriz,   label);                   n++;
  XtSetArg(args[n], XtNfromVert,    view);                    n++;
  XtSetArg(args[n], XtNtop,         XtChainBottom);           n++;
  XtSetArg(args[n], XtNbottom,      XtChainBottom);           n++;
  min = XtCreateManagedWidget(buf+1, asciiTextWidgetClass,
                                 form, args, n);
  if (!min) return -1;        /* create minimum widget */

  /* --- create input field for maximum --- */
  n = 0;                      /* build argument list */
  XtSetArg(args[n], XtNborderWidth, 0);                       n++;
  XtSetArg(args[n], XtNfromVert,    min);                     n++;
  if (left) { XtSetArg(args[n], XtNfromHoriz,   left);        n++; }
  else      { XtSetArg(args[n], XtNleft,        XtChainLeft); n++; }
  XtSetArg(args[n], XtNtop,         XtChainBottom);           n++;
  XtSetArg(args[n], XtNbottom,      XtChainBottom);           n++;
  XtSetArg(args[n], XtNjustify,     XtJustifyLeft);           n++;
  sprintf(buf, "l%s_max", name); /* create the widget name */
  label = XtCreateManagedWidget(buf, labelWidgetClass, form, args, n);
  if (!label) return -1;      /* create h_max label widget */
  n = 0;                      /* build argument list */
  XtSetArg(args[n], XtNborderWidth, 1);                       n++;
  XtSetArg(args[n], XtNeditType,    XawtextEdit);             n++;
  XtSetArg(args[n], XtNfromHoriz,   label);                   n++;
  XtSetArg(args[n], XtNfromVert,    min);                     n++;
  XtSetArg(args[n], XtNtop,         XtChainBottom);           n++;
  XtSetArg(args[n], XtNbottom,      XtChainBottom);           n++;
  max = XtCreateManagedWidget(buf+1, asciiTextWidgetClass,
                              form, args, n);
  if (!max) return -1;          /* create maximum widget */
  return 0;                     /* return 'ok' */
}  /* listbox() */

/*--------------------------------------------------------------------*/

static void atts (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Attributes... */
  static String empty = "";     /* an empty string */
  XtTranslations tla, tlo;      /* translations for text widgets */
  Widget    shell = NULL, form; /* widgets for dialog shell and form */
  Widget    lview, rview;       /* left and right viewport widget */
  Widget    lh_min, lh_max;     /* label widgets for horizontal */
  Widget    lv_min, lv_max;     /* and vertical minimum/maximum */
  Widget    ok, cancel;         /* widgets for ok and cancel buttons */
  Arg       args[16];           /* widget argument list */
  Cardinal  n;                  /* widget argument counter */
  Position  x,  y;              /* position of created dialog */
  Dimension wd, ht;             /* width and height of created dialog */
  Dimension lblwd, t;           /* label width and temp. buffer */
  int       dd;                 /* default distance of dialog items */

  if (!db_atts) {               /* if dialog box is not created */
    shell = XtCreatePopupShell("attDlg", transientShellWidgetClass,
              w_top, NULL, 0);  /* create a dialog box shell */
    if (!shell) { error(E_DIALOG); return; }
    XtSetArg(args[0], XtNborderWidth, 0);
    form = XtCreateManagedWidget("form", formWidgetClass,
             shell, args, 1);   /* create a form for the box layout */
    if (!form) { XtDestroyWidget(shell); error(E_DIALOG); return; }

    /* --- create listbox for horizontal axis --- */
    if (listbox(form, NULL,  "h") != 0) {
      XtDestroyWidget(shell); error(E_DIALOG); return; }
    lview    = XtNameToWidget(form, "*vh");
    di_h_att = XtNameToWidget(form, "*h_att");
    di_h_min = XtNameToWidget(form, "*h_min");
    di_h_max = XtNameToWidget(form, "*h_max");
    XtAddCallback(di_h_att, XtNcallback, lcb_att, (void*)0);

    /* --- create listbox for vertical axis --- */
    if (listbox(form, lview, "v") != 0) {
      XtDestroyWidget(shell); error(E_DIALOG); return; }
    rview    = XtNameToWidget(form, "*vv");
    di_v_att = XtNameToWidget(form, "*v_att");
    di_v_min = XtNameToWidget(form, "*v_min");
    di_v_max = XtNameToWidget(form, "*v_max");
    XtAddCallback(di_v_att, XtNcallback, lcb_att, (void*)1);

    /* --- create ok button --- */
    n = 0;                      /* build argument list */
    XtSetArg(args[n], XtNfromVert,    di_h_max);      n++;
    XtSetArg(args[n], XtNleft,        XtChainLeft);   n++;
    XtSetArg(args[n], XtNtop,         XtChainBottom); n++;
    XtSetArg(args[n], XtNbottom,      XtChainBottom); n++;
    ok = XtCreateManagedWidget("ok", commandWidgetClass,
           form, args, n);      /* create 'ok' button */
    if (!ok) { XtDestroyWidget(shell); error(E_DIALOG); return; }
    XtAddCallback(ok, XtNcallback, dcb_atts, (void*)1);

    /* --- create cancel button --- */
    n = 0;                      /* build argument list */
    XtSetArg(args[n], XtNfromHoriz,   ok);            n++;
    XtSetArg(args[n], XtNfromVert,    di_v_max);      n++;
    XtSetArg(args[n], XtNtop,         XtChainBottom); n++;
    XtSetArg(args[n], XtNright,       XtChainRight);  n++;
    XtSetArg(args[n], XtNbottom,      XtChainBottom); n++;
    cancel = XtCreateManagedWidget("cancel", commandWidgetClass,
               form, args, n);  /* create 'cancel' button */
    if (!cancel) { XtDestroyWidget(shell); error(E_DIALOG); return; }
    XtAddCallback(cancel, XtNcallback, dcb_atts, (void*)0);

    /* --- adapt widths of labels and gadgets --- */
    XtSetArg(args[0], XtNwidth, &t); lblwd = 1;
    lh_min = XtNameToWidget(form, "*lh_min");
    XtGetValues(lh_min, args, 1); if (t > lblwd) lblwd = t;
    lh_max = XtNameToWidget(form, "*lh_max");
    XtGetValues(lh_max, args, 1); if (t > lblwd) lblwd = t;
    lv_min = XtNameToWidget(form, "*lv_min");
    XtGetValues(lv_min, args, 1); if (t > lblwd) lblwd = t;
    lv_max = XtNameToWidget(form, "*lv_max");
    XtGetValues(lv_max, args, 1); if (t > lblwd) lblwd = t;
    XtVaSetValues(lh_min,   XtNwidth, lblwd, NULL);
    XtVaSetValues(lh_max,   XtNwidth, lblwd, NULL);
    XtVaSetValues(lv_min,   XtNwidth, lblwd, NULL);
    XtVaSetValues(lv_max,   XtNwidth, lblwd, NULL);
    XtVaGetValues(di_h_min, XtNwidth, &wd, NULL);
    #ifdef LINUX                /* if Linux Athena Widgets */
    XtVaGetValues(form,     XtNdefaultDistance, &dd, NULL);
    #else                       /* get the default distance, */
    dd = 4;                     /* if Sun or HP Athena Widgets, */
    #endif                      /* set the default distance */
    wd = (Dimension)(wd +lblwd +dd);  /* compute the listbox width */
    if (wd < LISTWD) wd = LISTWD;     /* and set it */
    XtVaSetValues(rview,    XtNwidth, wd, NULL);
    XtVaSetValues(lview,    XtNwidth, wd, NULL);
    XtVaSetValues(ok,       XtNwidth, wd, NULL);
    XtVaSetValues(cancel,   XtNwidth, wd, NULL);
    wd = (Dimension)(wd -lblwd -dd);  /* compute input field width */
    XtVaSetValues(di_h_min, XtNwidth, wd, NULL);
    XtVaSetValues(di_h_max, XtNwidth, wd, NULL);
    XtVaSetValues(di_v_min, XtNwidth, wd, NULL);
    XtVaSetValues(di_v_max, XtNwidth, wd, NULL);

    /* --- install translations --- */
    tla = XtParseTranslationTable(
            "<Btn1Down>,<Btn1Up>: Set()Notify()\n"
            "<Btn1Down>(2+):      dclick()");
    XtAugmentTranslations(di_h_att, tla);
    XtAugmentTranslations(di_v_att, tla);
    tla = XtParseTranslationTable(
            "<Key>Home  : beginning-of-line()\n"
            "<Key>End   : end-of-line()\n");
    tlo = XtParseTranslationTable(
            "Shift<Key>Tab   : db_next(-1)\n"
            "<Key>Tab        : db_next(+1)\n"
            "Shift<Key>Return: db_close(ok)\n"
            "<Key>Return     : db_close(apply)\n"
            "<Key>Escape     : db_close(cancel)\n"
            "!Alt<Key>o      : db_close(ok)\n"
            "!Alt<Key>c      : db_close(cancel)\n"
            "<Btn1Down>      : db_focus()\n");
    XtAugmentTranslations (di_h_min, tla);
    XtOverrideTranslations(di_h_min, tlo);
    XtAugmentTranslations (di_h_max, tla);
    XtOverrideTranslations(di_h_max, tlo);
    XtAugmentTranslations (di_v_min, tla);
    XtOverrideTranslations(di_v_min, tlo);
    XtAugmentTranslations (di_v_max, tla);
    XtOverrideTranslations(di_v_max, tlo);

    XtInstallAllAccelerators(form, form);
    db_atts = shell;            /* install actions and accelerators */
  }                             /* and note the created dialog box */
  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */

  /* --- fill dialog box fields --- */
  if (!names                    /* if the name list is not valid */
  ||  (attcnt <= 0)) {          /* or there are no num. attributes */
    XawListChange(di_h_att, (const char**)&empty, 1, 0, 1);
    XawListChange(di_v_att, (const char**)&empty, 1, 0, 1);
    XtSetArg(args[0], XtNstring, "");
    XtSetValues(di_h_min, args, 1);  /* clear the list boxes */
    XtSetValues(di_h_max, args, 1);  /* and the input gadgets */
    XtSetValues(di_v_min, args, 1);
    XtSetValues(di_v_max, args, 1); }
  else {                        /* if the name list is valid */
    XawListChange(di_h_att, (const char**)names, attcnt, 0, 1);
    XawListHighlight(di_h_att, selatt.h_att);
    XawListChange(di_v_att, (const char**)names, attcnt, 0, 1);
    XawListHighlight(di_v_att, selatt.v_att);
    sprintf(buf, "%g", selatt.h_rng.min);
    XtVaSetValues(di_h_min, XtNstring, buf, NULL);
    sprintf(buf, "%g", selatt.h_rng.max);
    XtVaSetValues(di_h_max, XtNstring, buf, NULL);
    sprintf(buf, "%g", selatt.v_rng.min);
    XtVaSetValues(di_v_min, XtNstring, buf, NULL);
    sprintf(buf, "%g", selatt.v_rng.max);
    XtVaSetValues(di_v_max, XtNstring, buf, NULL);
  }                             /* set attribute names and ranges */
  XtVaSetValues(XtParent(di_h_att), XtNheight, LISTHT, NULL);
  XtVaSetValues(XtParent(di_v_att), XtNheight, LISTHT, NULL);

  /* --- show dialog box --- */
  XtVaGetValues(w_top,   XtNx, &x, XtNy, &y, NULL);
  XtVaSetValues(db_atts, XtNx,  x, XtNy,  y, NULL);
  XtPopup(db_atts, XtGrabExclusive);   /* place and show dialog box */
  XtSetKeyboardFocus(db_atts, di_h_min);  /* and set keyboard focus */

  if (shell) {                  /* if dialog box has been created */
    XtVaGetValues(shell, XtNwidth,   &wd, XtNheight,   &ht, NULL);
    XtVaSetValues(shell, XtNminWidth, wd, XtNminHeight, ht, NULL);
    XtOverrideTranslations(shell,
      XtParseTranslationTable("<Message>WM_PROTOCOLS: db_close()"));
    XSetWMProtocols(XtDisplay(shell), XtWindow(shell),
      &wm_delwin, 1);           /* override window manager delete */
  }
}  /* atts() */

/*--------------------------------------------------------------------*/

static void colors (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Colors... */
  char buf[32];                 /* ditto, for texts */

  if (!psh_colors) {            /* if property sheet not created yet */
    psh_colors = psh_create("colorsDlg", w_top, 1);
    if (!psh_colors) return;     /* create a property sheet */
    if ((psh_addline(psh_colors, "lmode")               != 0)
    ||  (psh_additem(psh_colors, "mode",   PSH_MENU, 0) != 0)
    ||  (psh_menuadd(psh_colors, "mode",   "color")     != 0)
    ||  (psh_menuadd(psh_colors, "mode",   "grey")      != 0)
    ||  (psh_addline(psh_colors, "lcoloff")             != 0)
    ||  (psh_additem(psh_colors, "coloff", PSH_EDIT, 0) != 0)
    ||  (psh_addline(psh_colors, "lbase")               != 0)
    ||  (psh_additem(psh_colors, "base",   PSH_MENU, 0) != 0)
    ||  (psh_menuadd(psh_colors, "base",   "black")     != 0)
    ||  (psh_menuadd(psh_colors, "base",   "white")     != 0)
    ||  (psh_addline(psh_colors, "lshdcnt")             != 0)
    ||  (psh_additem(psh_colors, "shdcnt", PSH_EDIT, 0) != 0)) {
      psh_delete(psh_colors); error(E_DIALOG); return; }
  }                             /* add dialog items */
  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  psh_setval(psh_colors, "mode",         /* set shading mode */
    (const char*)(ptrdiff_t)((parblk.colcnt > 0) ? 0 : 1));
  sprintf(buf, "%g", parblk.coloff);
  psh_setval(psh_colors, "coloff", buf); /* set color offset */
  psh_setval(psh_colors, "base",         /* set shade base */
    (const char*)(ptrdiff_t)((parblk.shdbase == SHD_BLACK) ? 0 : 1));
  sprintf(buf, "%d", parblk.shdcnt);
  psh_setval(psh_colors, "shdcnt", buf); /* set number of shades */
  psh_handle(psh_colors, dcb_colors);    /* and handle dialog */
}  /* colors() */

/*--------------------------------------------------------------------*/

static void marks (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Parameters... */
  char buf[32];                 /* ditto, for texts */

  if (!psh_marks) {             /* if property sheet not created yet */
    psh_marks = psh_create("marksDlg", w_top, 1);
    if (!psh_marks) return;     /* create a property sheet */
    if ((psh_addline(psh_marks, "lsize")               != 0)
    ||  (psh_additem(psh_marks, "size",   PSH_EDIT, 0) != 0)
    ||  (psh_addline(psh_marks, "lborder")             != 0)
    ||  (psh_additem(psh_marks, "border", PSH_MENU, 0) != 0)
    ||  (psh_menuadd(psh_marks, "border", "black")     != 0)
    ||  (psh_menuadd(psh_marks, "border", "white")     != 0)) {
      psh_delete(psh_marks); error(E_DIALOG); return; }
  }                             /* add dialog items */
  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  sprintf(buf, "%d", parblk.size);
  psh_setval(psh_marks, "size", buf);  /* set marker size and color */
  psh_setval(psh_marks, "border",(const char*)(ptrdiff_t)parblk.border);
  psh_handle(psh_marks, dcb_marks);    /* and handle dialog */
}  /* marks() */

/*--------------------------------------------------------------------*/

static void ngbors (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Neighbors... */
  char buf[32];                 /* ditto, for texts */

  if (!psh_ngbors) {            /* if property sheet not created yet */
    psh_ngbors = psh_create("ngborsDlg", w_top, 1);
    if (!psh_ngbors) return;    /* create a property sheet */
    if ((psh_addline(psh_ngbors, "lnbcnt")              != 0)
    ||  (psh_additem(psh_ngbors, "nbcnt",  PSH_EDIT, 0) != 0)
    ||  (psh_addline(psh_ngbors, "lgraph")              != 0)
    ||  (psh_additem(psh_ngbors, "graph",  PSH_MENU, 0) != 0)
    ||  (psh_menuadd(psh_ngbors, "graph",  "no")        != 0)
    ||  (psh_menuadd(psh_ngbors, "graph",  "yes")       != 0)) {
      psh_delete(psh_ngbors); error(E_DIALOG); return; }
  }                             /* add dialog items */
  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  sprintf(buf, "%d", parblk.nbcnt);
  psh_setval(psh_ngbors, "nbcnt", buf);  /* set number of neighbors */
  psh_setval(psh_ngbors, "graph",        /* set neighbor graph flag */
            (const char*)(ptrdiff_t)parblk.graph);
  psh_handle(psh_ngbors, dcb_ngbors);    /* handle dialog */
}  /* ngbors() */

/*--------------------------------------------------------------------*/

static void redraw (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Redraw */
  static Dimension owd = 0;     /* old window width */
  static Dimension oht = 0;     /* old window height */
  int        i, k, n;           /* loop variables, buffers */
  int        clscnt;            /* number ofclusters */
  int        colcnt;            /* number of colors */
  int        tplcnt;            /* number of tuples */
  Dimension  wd, ht;            /* window extensions */
  XRectangle rc;                /* clipping rectangle */
  int        xid, yid;          /* attribute identifiers */
  int        x, y, x_prev;      /* pixel coordinates */
  int        *xcs, *ycs;        /* additional pixel coordinates */
  double     xoff, yoff;        /* offsets of window borders */
  double     xscl, yscl;        /* scaling factors */
  double     *msds;             /* membership degrees */
  double     val, sum, max;     /* function value, sum and maximum */
  PIXEL      pxl, pxl_prev;     /* current and previous pixel value */
  PIXEL      mrk;               /* pixel value for markers */
  int        col, shd, shdcnt;  /* color, shade, and number of shades */
  XEvent     de;                /* dummy event */
  double     a, b;              /* transformation constants */
  INST       *xinst, *yinst;    /* attribute instances */
  ATT        *att;              /* to traverse the attributes */
  TUPLE      *tpl;              /* to traverse the data tuples */
  float      f;                 /* temporary buffer */
  char       *msg;              /* buffer for message */

  /* --- initialize graphics --- */
  if (!gc) {                    /* if graphics context not created */
    gc = XCreateGC(display, window, 0, NULL);
    if (!gc) { error(E_GRAPH); return; }
  }                             /* create a graphics context */
  if (!ptlset || !ptltab) {     /* if there is no cluster set */
    XClearWindow(display, window);
    msg = (ptlset) ? "Load a data table." : "Load a cluster set.";
    XDrawString(display, window, gc, 8, 16, msg, (int)strlen(msg));
    return;                     /* clear window, draw message, */
  }                             /* and abort */
  XtVaGetValues(w_view, XtNwidth, &wd, XtNheight, &ht, NULL);
  if (e && (e->type == Expose)  /* if the window was exposed, */
  &&  (wd == owd) && (ht == oht)) {        /* but not resized */
    rc.x      = (short)e->xexpose.x;    /* build and set */
    rc.y      = (short)e->xexpose.y;    /* clipping rectangle */
    rc.width  = (unsigned short)e->xexpose.width;
    rc.height = (unsigned short)e->xexpose.height;
    XSetClipRectangles(display, gc, 0, 0, &rc, 1, Unsorted); }
  else {                        /* if the window was resized */
    rc.x = rc.y = 0;            /* or this is a forced redraw, */
    rc.width  = wd;             /* redraw all of the window */
    rc.height = ht;             /* (set full window rectangle) */
    while (XCheckTypedWindowEvent(display, window, Expose, &de));
  }                             /* remove all exposure events */
  owd = wd; oht = ht;           /* note (new) window dimensions */

  /* --- initialize parameters --- */
  for (i = as_attcnt(attset); --i >= 0; ) {
    att = as_att(attset, i);    /* traverse the attributes */
    if      (i == selatt.h_att) att_setmark(att,  0);
    else if (i == selatt.v_att) att_setmark(att,  1);
    else                        att_setmark(att, -1);
  }                             /* mark the visible attributes */
  xid    = map_fwd[selatt.h_att];  /* get attribute identifiers */
  yid    = map_fwd[selatt.v_att];  /* and instances */
  xinst  = att_inst(as_att(attset, xid));
  yinst  = att_inst(as_att(attset, yid));
  xoff   = ranges[selatt.h_att].min; /* get coordinate offsets and */
  yoff   = ranges[selatt.v_att].max; /* compute the scaling factors */
  xscl   = (ranges[selatt.h_att].max -xoff) /(wd -1);
  yscl   = (yoff -ranges[selatt.v_att].min) /(ht -1);
  clscnt = ptl_clscnt(ptlset);  /* get the number of clusters */
  tplcnt = tab_tplcnt(ptltab);  /* and the number of tuples */
  colcnt = shd_colcnt(shades);  /* get the number of colors */
  shdcnt = shd_shdcnt(shades);  /* and the number of shades */
  a      = 1.0/clscnt;          /* compute transformation constants */
  b      = (clscnt > 1) ? clscnt/(clscnt-1.0) : 1;
  msds   = dists +tplcnt;       /* get vector of membership degrees */
  pxl_prev = shd_pixel(shades, 0, 0);

  /* --- draw rectangle contents --- */
  if (ptl_getnorm(ptlset)) {    /* if normalized data */
    xoff = xoff *scales[xid] +offs[xid]; xscl *= scales[xid];
    yoff = yoff *scales[yid] +offs[yid]; yscl *= scales[yid];
  }                             /* adapt the scaling parameters */
  for (y = rc.y; y < rc.y +rc.height; y++) {
    for (x = x_prev = rc.x +rc.width; --x >= rc.x; ) {
      xinst->f = (float)(xoff +x *xscl); /* compute real coordinates */
      yinst->f = (float)(yoff -y *yscl); /* of a window pixel */
      dv_fromtab(dists, tplcnt, nbids, parblk.nbcnt, ptltab, NULL);
      ptl_exec(ptlset, dists, nbids, parblk.nbcnt, msds);
      max = sum = 0;            /* execute the cluster set, */
      col = 0;                  /* then traverse the clusters */
      for (i = clscnt; --i >= 0; ) {
        if (msds[i] > max) max = msds[col = i];
        sum += msds[i];         /* find the maximal membership */
      }                         /* and sum the membership degrees */
      val = (sum > 0) ? (max/sum -a) *b : 0;
      shd = (int)(shdcnt *val); /* compute the color shade */
      if (shd    >= shdcnt) shd = shdcnt-1;
      if (shd    <  0)      shd = 0;
      if (colcnt <= 0)      col = 0;
      pxl = shd_pixel(shades, col, shd);
      if (pxl == pxl_prev) continue;
      XSetForeground(display, gc, pxl_prev);
      XDrawLine(display, window, gc, x_prev, y, x, y);
      x_prev   = x;             /* draw a line segment and note */
      pxl_prev = pxl;           /* the end of the line segment */
    }                           /* and the new pixel value */
    XSetForeground(display, gc, pxl_prev);
    XDrawLine(display, window, gc, x_prev, y, x, y);
  }                             /* draw last line segment */

  /* --- compute data point coordinates --- */
  xoff = ranges[selatt.h_att].min; /* get coordinate offsets and */
  yoff = ranges[selatt.v_att].max; /* compute the scaling factors */
  xscl = (ranges[selatt.h_att].max -xoff) /(wd -1);
  yscl = (yoff -ranges[selatt.v_att].min) /(ht -1);
  xcs = nbids; ycs = nbids +(k = tab_tplcnt(table));
  while (--k >= 0) {            /* get the coordinate vectors and */
    tpl = tab_tpl(table, k);    /* traverse the data points */
    f = tpl_colval(tpl, xid)->f;
    if (isnan(f)) { xcs[k] = INT_MIN; continue; }
    xcs[k] = (int)((f -xoff) /xscl +0.5);
    f = tpl_colval(tpl, yid)->f;
    if (isnan(f)) { xcs[k] = INT_MIN; continue; }
    ycs[k] = (int)((yoff -f) /yscl +0.5);
  }                             /* compute the screen coordinates */

  /* --- draw neighborhood graph --- */
  mrk = (parblk.border == 0) ? black : white;
  if (parblk.graph) {           /* if to draw a neighborhood graph */
    XSetForeground(display, gc, mrk);
    for (k = tab_tplcnt(table); --k >= 0; ) {
      if (xcs[k] <= INT_MIN) continue;
      x = xcs[k]; y = ycs[k];   /* traverse the data points */
      for (n = (distmat) ? parblk.nbcnt : k; --n >= 0; ) {
        if (distmat) dm_getdist(distmat, k, n, &i);
        else         i = n;     /* traverse the neighbors */
        if (xcs[i] <= INT_MIN) continue;
        XDrawLine(display, window, gc, x, y, xcs[i], ycs[i]);
      }                         /* connect a data point */
    }                           /* to all its neighbors */
  }                             /* (draw neighborhood graph) */

  /* --- draw data --- */
  n = parblk.size;              /* get the marker parameters */
  i = n >> 1;                   /* (size and offset) */
  for (k = tab_tplcnt(table); --k >= 0; ) {
    if (xcs[k] <= INT_MIN) continue;
    x = xcs[k] -i; y = ycs[k] -i;
    XSetForeground(display, gc, grey);
    XFillArc(display, window, gc, x, y, (unsigned)n, (unsigned)n,
             0, 360*64);
    XSetForeground(display, gc, mrk);
    XDrawArc(display, window, gc, x, y, (unsigned)n, (unsigned)n,
             0, 360*64);
  }                             /* draw a circle at the data point */

  XFlush(display);              /* flush buffered drawing commands */
  XSetClipMask(display, gc, None);      /* and clear clipping mask */
}  /* redraw() */

/*--------------------------------------------------------------------*/

static void about (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- 'About' dialog box */
  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  if (db_about(w_top, 0, dcb_nop) != 0)
    error(E_DIALOG);            /* show 'about' dialog box */
}  /* about() */

/*----------------------------------------------------------------------
  Main Function
----------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main program */
  Arg      args[8];             /* widget arguments */
  Cardinal n;                   /* argument counter */
  Colormap colmap;              /* color map of display */
  XColor   xc[2];               /* buffer for colors */

  XtSetLanguageProc(NULL, (XtLanguageProc)NULL, NULL);

  /* --- create main window --- */
  w_top = XtAppInitialize(&appctx, "XPtlView",
            NULL, 0, &argc, argv, fallback, NULL, 0);
  if (!w_top) error(E_INIT);    /* create toplevel widget */
  n = 0;                        /* build argument list */
  XtSetArg(args[n], XtNdefaultDistance, 0);    n++;
  XtSetArg(args[n], XtNresizable,       True); n++;
  w_main = XtCreateManagedWidget("main", formWidgetClass,
             w_top, args, n);   /* create main widget */
  if (!w_main) error(E_WIDGET); /* (form for window layout) */
  menu = mn_create(w_main);     /* create a menu bar and */
  if (!menu) error(E_NOMEM);    /* add titles and items */
  if ((mn_addtitle(menu, "file",      0, 0,          0) != 0)
  ||  (mn_additem (menu, "ld_ptlset", 0, mcb_ptlset, 0) != 0)
  ||  (mn_additem (menu, "ld_data",   0, mcb_data,   0) != 0)
  ||  (mn_additem (menu, "format",    0, mcb_format, 0) != 0)
  ||  (mn_additem (menu, NULL,        0, 0,          0) != 0)
  ||  (mn_additem (menu, "quit",      0, mcb_quit,   0) != 0)
  ||  (mn_addtitle(menu, "actions",   0, 0,          0) != 0)
  ||  (mn_additem (menu, "atts",      0, mcb_atts,   0) != 0)
  ||  (mn_additem (menu, "colors",    0, mcb_colors, 0) != 0)
  ||  (mn_additem (menu, "marks",     0, mcb_marks,  0) != 0)
  ||  (mn_additem (menu, "ngbors",    0, mcb_ngbors, 0) != 0)
  ||  (mn_additem (menu, NULL,        0, 0,          0) != 0)
  ||  (mn_additem (menu, "redraw",    0, mcb_redraw, 0) != 0)
  ||  (mn_addtitle(menu, "help",      0, 0,          0) != 0)
  ||  (mn_additem (menu, "about",     0, mcb_about,  0) != 0)
  ||  (mn_addtitle(menu, NULL,        0, 0,          0) != 0))
    error(E_WIDGET);
  mn_resize(menu, INITXEXT);    /* set width of menu bar */

  /* --- create viewport --- */
  n = 0;                        /* build argument list */
  XtSetArg(args[n], XtNfromVert,  mn_menubar(menu)); n++;
  XtSetArg(args[n], XtNleft,      XtChainLeft);      n++;
  XtSetArg(args[n], XtNtop,       XtChainTop);       n++;
  XtSetArg(args[n], XtNright,     XtChainRight);     n++;
  XtSetArg(args[n], XtNbottom,    XtChainBottom);    n++;
  XtSetArg(args[n], XtNresizable, True);             n++;
  w_view = XtCreateManagedWidget("view",
             coreWidgetClass, w_main, args, n);
  if (!w_view) error(E_WIDGET); /* create viewport */
  XtVaSetValues(w_view, XtNwidth,     INITXEXT,
                        XtNheight,    INITYEXT, NULL);
  XtVaSetValues(w_top,  XtNminWidth,  MINXEXT,
                        XtNminHeight, MINYEXT,  NULL);
                                /* set window size */
  XtOverrideTranslations(w_view,/* install redraw function */
    XtParseTranslationTable("<Expose>: redraw()"));

  /* --- install actions, accelerators and translations --- */
  XtAppAddActions(appctx, actions, XtNumber(actions));
  XtInstallAllAccelerators(w_main, w_main);

  /* --- show window --- */
  XtRealizeWidget(w_top);       /* show window on display */
  display = XtDisplay(w_top);   /* get display of top window */
  window  = XtWindow(w_view);   /* and drawable for window */

  if (db_alert(w_top, dcb_nop, NULL) != 0)
    error(E_WIDGET);            /* create alert dialog box */

  /* --- override window manager delete --- */
  XtOverrideTranslations(w_top, /* override window manager delete */
    XtParseTranslationTable("<Message>WM_PROTOCOLS: quit()"));
  wm_delwin = XInternAtom(XtDisplay(w_top), "WM_DELETE_WINDOW", False);
  XSetWMProtocols(XtDisplay(w_top), XtWindow(w_top), &wm_delwin, 1);

  /* --- allocate colors --- */
  screen = DefaultScreen(display);
  colmap = DefaultColormap(display, screen);
  if (XAllocNamedColor(display, colmap, "grey", xc+1, xc) == 0)
    error(E_COLOR);             /* allocate colors */
  black  = BlackPixel(display, screen);
  white  = WhitePixel(display, screen);
  grey   = xc[1].pixel;         /* enter the color values */

  /* --- evaluate command line arguments --- */
  if (argc > 1) dcb_ptlset(NULL, (XtPointer)2, (XtPointer)argv[1]);
  if (argc > 2) dcb_data  (NULL, (XtPointer)1, (XtPointer)argv[2]);

  /* --- start event loop --- */
  XtAppMainLoop(appctx);        /* enter main event loop */
  return 0;                     /* return 'ok' */
}  /* main() */
